from click import group, version_option

from firebolt_cli import __version__
from firebolt_cli.configure import configure
from firebolt_cli.database import database
from firebolt_cli.engine import engine
from firebolt_cli.ingest import ingest
from firebolt_cli.query import query
from firebolt_cli.table import table
from firebolt_cli.utils import construct_shortcuts


@group(
    cls=construct_shortcuts(
        shortages={
            "config": "configure (config)",
            "configure": "configure (config)",
            "db": "database (db)",
            "database": "database (db)",
            "tb": "table (tb)",
            "table": "table (tb)",
        }
    )
)
@version_option(__version__, "-V", "--version")
def main() -> None:
    """
    Firebolt command line utility.
    """


main.add_command(ingest)
main.add_command(configure)
main.add_command(engine)
main.add_command(query)
main.add_command(database)
main.add_command(table)



import os
import sys
from typing import Dict

import click
from click import Choice, command, echo, option
from firebolt.common.exception import FireboltError
from firebolt_ingest.table_model import Table
from firebolt_ingest.table_service import TableService

from firebolt_cli.common_options import (
    common_options,
    default_from_config_file,
)
from firebolt_cli.utils import (
    create_connection,
    exit_on_firebolt_exception,
    read_from_file,
)


@command()
@common_options
@option(
    "--engine-name",
    help="Name or url of the engine to use for SQL queries",
    envvar="FIREBOLT_ENGINE_NAME",
    callback=default_from_config_file(required=True),
)
@option(
    "--database-name",
    help="Database name to use for SQL queries",
    envvar="FIREBOLT_DATABASE_NAME",
    callback=default_from_config_file(required=True),
)
@option(
    "--file",
    help="Path to the yaml config file",
    type=click.Path(exists=True),
    required=True,
)
@option(
    "--mode",
    help="Mode of ingestion. "
    "Overwrite (default) - will drop and fully recreate fact table."
    "Append - will only insert files, that aren't in the fact table yet.",
    required=False,
    default="overwrite",
    type=Choice(["append", "overwrite"], case_sensitive=False),
)
@option(
    "--firebolt_dont_wait_for_upload_to_s3",
    help="Don't wait for upload part to S3 on insert query finish. ",
    is_flag=True,
    required=False,
    default=False,
)
@option(
    "--advanced_mode",
    help="execute set advanced_mode=1",
    is_flag=True,
    required=False,
    default=False,
)
@option(
    "--use_short_column_path_parquet",
    help="Use short parquet column path "
    "(skipping repeated nodes and their child node).",
    is_flag=True,
    required=False,
    default=False,
)
@exit_on_firebolt_exception
def ingest(**raw_config_options: str) -> None:
    """
    [Beta] Ingest the data from external to fact table.
    """

    table = Table.parse_yaml(read_from_file(raw_config_options["file"]))

    with create_connection(**raw_config_options) as connection:
        ts = TableService(table, connection)

        params: Dict[str, bool] = {
            "firebolt_dont_wait_for_upload_to_s3": bool(
                raw_config_options["firebolt_dont_wait_for_upload_to_s3"]
            ),
            "advanced_mode": bool(raw_config_options["advanced_mode"]),
            "use_short_column_path_parquet": bool(
                raw_config_options["use_short_column_path_parquet"]
            ),
        }

        if raw_config_options["mode"] == "overwrite":
            ts.insert_full_overwrite(**params)
        elif raw_config_options["mode"] == "append":
            ts.insert_incremental_append(**params)
        else:
            raise FireboltError(f"Mode: {raw_config_options['mode']} unknown.")

        echo(
            f"Ingestion from 'ex_{table.table_name}' "
            f"to '{table.table_name}' was successful."
        )

        if not ts.verify_ingestion():
            echo(
                "WARNING: Nevertheless some discrepancy between fact and "
                "external table were found. It is recommended to do the full "
                "overwrite of the fact table.",
                err=True,
            )
            sys.exit(os.EX_DATAERR)



import os
import sys
from datetime import timedelta
from typing import Callable, Optional

from click import (
    Choice,
    IntRange,
    argument,
    command,
    confirm,
    echo,
    group,
    option,
)
from firebolt.common.exception import FireboltError
from firebolt.model.engine import Engine
from firebolt.service.manager import ResourceManager
from firebolt.service.types import (
    EngineStatusSummary,
    EngineType,
    WarmupMethod,
)

from firebolt_cli.common_options import (
    common_options,
    default_from_config_file,
    json_option,
)
from firebolt_cli.utils import (
    construct_resource_manager,
    construct_shortcuts,
    convert_bytes,
    exit_on_firebolt_exception,
    get_default_database_engine,
    prepare_execution_result_line,
    prepare_execution_result_table,
)


@group(
    cls=construct_shortcuts(
        shortages={
            "list": "list (ls)",
            "ls": "list (ls)",
        }
    )
)
def engine() -> None:
    """
    Manage engines.
    """


def get_engine_from_name_or_default(
    rm: ResourceManager, engine_name: Optional[str], database_name: Optional[str]
) -> Engine:
    """
    Returns engine either from its name, or a default engine deducted
    from database_name. At least one engine_name or database_name should
    be provided, raises an Error otherwise.
    """
    if engine_name is not None:
        return rm.engines.get_by_name(name=engine_name)
    elif database_name is not None:
        return get_default_database_engine(rm, database_name)
    else:
        raise FireboltError("Either engine name or database name has to be specified")


def start_stop_generic(
    engine: Engine,
    action: str,
    accepted_initial_states: set,
    accepted_final_states: set,
    accepted_final_nowait_states: set,
    wrong_initial_state_error: str,
    failure_message: str,
    success_message: str,
    success_message_nowait: str,
    **raw_config_options: str,
) -> None:

    current_status_name = (
        engine.current_status_summary.name
        if engine.current_status_summary
        else EngineStatusSummary.ENGINE_STATUS_SUMMARY_UNSPECIFIED.name
    )

    if engine.current_status_summary not in accepted_initial_states:
        raise FireboltError(
            wrong_initial_state_error.format(
                name=engine.name,
                state=current_status_name,
            )
        )

    if action == "start":
        engine = engine.start(wait_for_startup=raw_config_options["wait"])
    elif action == "stop":
        engine = engine.stop(wait_for_stop=raw_config_options["wait"])
    elif action == "restart":
        engine = engine.restart(wait_for_startup=raw_config_options["wait"])
    else:
        assert False, "not available action"

    if (
        engine.current_status_summary in accepted_final_nowait_states
        and not raw_config_options["wait"]
    ):
        echo(success_message_nowait.format(name=engine.name))
    elif engine.current_status_summary in accepted_final_states:
        echo(success_message.format(name=engine.name))
    else:
        raise FireboltError(
            failure_message.format(name=engine.name, status=current_status_name)
        )


@command()
@common_options
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Alternatively to engine name, database name could be specified, "
    "its default engine will be used",
    hidden=True,
    callback=default_from_config_file(required=False),
)
@option(
    "--wait/--no-wait",
    help="Wait until the engine is started.",
    is_flag=True,
    default=False,
)
@argument("engine_name", type=str, required=False)
@exit_on_firebolt_exception
def start(**raw_config_options: str) -> None:
    """
    Start an existing ENGINE_NAME. If ENGINE_NAME is not set,
    uses default engine instead.
    """

    rm = construct_resource_manager(**raw_config_options)
    engine = get_engine_from_name_or_default(
        rm, raw_config_options["engine_name"], raw_config_options["database_name"]
    )

    if (
        engine.current_status_summary
        == EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED
    ):
        raise FireboltError(
            f"Engine {engine.name} is in a failed state.\n"
            f"You need to restart an engine first:\n"
            f"$ firebolt engine restart {engine.name}"
        )

    start_stop_generic(
        engine=engine,
        action="start",
        accepted_initial_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING,
        },
        accepted_final_states={EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING},
        accepted_final_nowait_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STARTING
        },
        wrong_initial_state_error="Engine {name} is not in a stopped state."
        "The current engine state is {state}.",
        success_message="Engine {name} is successfully started.",
        success_message_nowait="Start request for engine {name} is successfully sent.",
        failure_message="Engine {name} failed to start. Engine status: {status}.",
        **raw_config_options,
    )


@command()
@common_options
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Alternatively to engine name, database name could be specified, "
    "its default engine will be used",
    hidden=True,
    callback=default_from_config_file(required=False),
)
@option(
    "--wait/--no-wait",
    help="Wait until the engine is stopped.",
    is_flag=True,
    default=False,
)
@argument("engine_name", type=str, required=False)
@exit_on_firebolt_exception
def stop(**raw_config_options: str) -> None:
    """
    Stop an existing ENGINE_NAME. If ENGINE_NAME is not set,
    uses default engine instead.
    """

    rm = construct_resource_manager(**raw_config_options)
    engine = get_engine_from_name_or_default(
        rm, raw_config_options["engine_name"], raw_config_options["database_name"]
    )

    start_stop_generic(
        engine=engine,
        action="stop",
        accepted_initial_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STARTING_INITIALIZING,
        },
        accepted_final_states={EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED},
        accepted_final_nowait_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPED,
        },
        wrong_initial_state_error="Engine {name} is not in a "
        "running or initializing state. The current engine state is {state}.",
        success_message="Engine {name} is successfully stopped.",
        success_message_nowait="Stop request for engine {name} is successfully sent.",
        failure_message="Engine {name} failed to stop. Engine status: {status}.",
        **raw_config_options,
    )


def engine_properties_options(create_mode: bool = True) -> Callable:
    """
    decorator for engine create/update common options

    :param create_mode: True for create, will make some options required
    """
    _ENGINE_OPTIONS = [
        option(
            "--name",
            help="Name of the engine.",
            type=str,
            required=True,
        ),
        option(
            "--spec",
            help="Engine spec. Run 'firebolt engine get-instance-types' "
            "to get a list of available spec",
            type=str,
            required=create_mode,
        ),
        option(
            "--description",
            help="Engine description (max: 64 characters).",
            type=str,
            default="" if create_mode else None,
            required=False,
        ),
        option(
            "--type",
            help='Engine type: "rw" for general purpose '
            'and "ro" for data analytics.',
            type=Choice(list(ENGINE_TYPES.keys()), case_sensitive=False),
            default="ro" if create_mode else None,
            required=False,
        ),
        option(
            "--scale",
            help="The number of engine nodes. Value entered must be between 1 and 128.",
            type=IntRange(1, 128, clamp=False),
            default=1 if create_mode else None,
            required=False,
            show_default=True,
            metavar="INTEGER",
        ),
        option(
            "--use-spot/--no-use-spot",
            help="Use spot instances",
            is_flag=True,
            default=None,
            required=False,
        ),
        option(
            "--auto-stop",
            help="Stop engine automatically after specified time in minutes."
            "Value entered must be between 0 and 43200"
            "(max value is equal to 30 days). "
            "Setting auto-stop to zero disables auto-stop.",
            type=IntRange(0, 30 * 24 * 60, clamp=False),
            default=20 if create_mode else None,
            required=False,
            show_default=True,
            metavar="INTEGER",
        ),
        option(
            "--warmup",
            help="Engine warmup method. "
            "Minimal(min), Preload indexes(ind), Preload all data(all)",
            type=Choice(list(WARMUP_METHODS.keys())),
            default="ind" if create_mode else None,
            required=False,
            show_default=True,
        ),
    ]

    def _engine_properties_options_inner(command: Callable) -> Callable:
        for add_option in reversed(_ENGINE_OPTIONS):
            command = add_option(command)
        return command

    return _engine_properties_options_inner


def echo_engine_information(
    rm: ResourceManager, engine: Engine, use_json: bool
) -> None:
    """

    :param engine:
    :param database:
    :param use_json:
    :return:
    """

    revision = None
    instance_type = None
    if engine.latest_revision_key:
        revision = rm.engine_revisions.get_by_key(engine.latest_revision_key)
        instance_type = rm.instance_types.instance_types_by_key[
            revision.specification.db_compute_instances_type_key
        ]

    def _format_auto_stop(auto_stop: str) -> str:
        """
        auto_stop could be set either 0 or to a value with ending with m or s
        if it is the case then we print its timedelta or "ALWAYS ON"
        if not the original auto_stop parameter is returned
        """
        val = int(auto_stop[:-1])

        if val == 0:
            return "ALWAYS ON"

        if auto_stop[-1] == "m":
            return str(timedelta(minutes=val))
        elif auto_stop[-1] == "s":
            return str(timedelta(seconds=val))
        else:
            return auto_stop

    echo(
        prepare_execution_result_line(
            data=[
                engine.name,
                engine.description,
                engine.current_status_summary.name
                if engine.current_status_summary
                else None,
                _format_auto_stop(engine.settings.auto_stop_delay_duration),
                revision.specification.db_compute_instances_use_spot
                if revision
                else "",
                engine.settings.preset,
                engine.settings.warm_up,
                str(engine.create_time),
                engine.database.name if engine.database else None,
                instance_type.name if instance_type else "",
                revision.specification.db_compute_instances_count if revision else "",
            ],
            header=[
                "name",
                "description",
                "status",
                "auto_stop",
                "is_spot_instance",
                "preset",
                "warm_up",
                "create_time",
                "attached_to_database",
                "instance_type",
                "scale",
            ],
            use_json=bool(use_json),
        )
    )


ENGINE_TYPES = {"rw": EngineType.GENERAL_PURPOSE, "ro": EngineType.DATA_ANALYTICS}
WARMUP_METHODS = {
    "min": WarmupMethod.MINIMAL,
    "ind": WarmupMethod.PRELOAD_INDEXES,
    "all": WarmupMethod.PRELOAD_ALL_DATA,
}


@command()
@common_options
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Alternatively to engine name, database name could be specified, "
    "its default engine will be used",
    hidden=True,
    callback=default_from_config_file(required=False),
)
@option(
    "--wait/--no-wait",
    help="Wait until the engine is restarted.",
    is_flag=True,
    default=False,
)
@argument("engine_name", type=str, required=False)
@exit_on_firebolt_exception
def restart(**raw_config_options: str) -> None:
    """
    Restart an existing ENGINE_NAME. If ENGINE_NAME is not set,
    uses default engine instead.
    """

    rm = construct_resource_manager(**raw_config_options)
    engine = get_engine_from_name_or_default(
        rm, raw_config_options["engine_name"], raw_config_options["database_name"]
    )

    start_stop_generic(
        engine=engine,
        action="restart",
        accepted_initial_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_FAILED,
        },
        accepted_final_states={EngineStatusSummary.ENGINE_STATUS_SUMMARY_RUNNING},
        accepted_final_nowait_states={
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STOPPING,
            EngineStatusSummary.ENGINE_STATUS_SUMMARY_STARTING,
        },
        wrong_initial_state_error="Engine {name} is not in a running or failed state."
        " The current engine state is {state}.",
        success_message="Engine {name} is successfully restarted.",
        success_message_nowait="Restart request for engine {name} "
        "is successfully sent.",
        failure_message="Engine {name} failed to restart. Engine status: {status}.",
        **raw_config_options,
    )


@command()
@common_options
@engine_properties_options(create_mode=True)
@option(
    "--database-name",
    help="Name of the database the engine should be attached to.",
    type=str,
    required=True,
)
@json_option
@exit_on_firebolt_exception
def create(**raw_config_options: str) -> None:
    """
    Creates engine with the requested parameters.
    """
    rm = construct_resource_manager(**raw_config_options)

    database = rm.databases.get_by_name(name=raw_config_options["database_name"])
    region = rm.regions.get_by_key(database.compute_region_key)

    engine = rm.engines.create(
        name=raw_config_options["name"],
        spec=raw_config_options["spec"],
        region=region.name,
        engine_type=ENGINE_TYPES[raw_config_options["type"]],
        scale=int(raw_config_options["scale"]),
        auto_stop=int(raw_config_options["auto_stop"]),
        warmup=WARMUP_METHODS[raw_config_options["warmup"]],
        description=raw_config_options["description"],
        revision_spec_kwargs={
            "db_compute_instances_use_spot": True
            if raw_config_options["use_spot"]
            else False
        },
    )

    try:
        database.attach_to_engine(engine=engine, is_default_engine=True)
    except (FireboltError, RuntimeError) as err:
        engine.delete()
        raise err

    if not raw_config_options["json"]:
        echo(
            f"Engine {engine.name} is successfully created "
            f"and attached to the {database.name}."
        )

    echo_engine_information(rm, engine, bool(raw_config_options["json"]))


@command()
@common_options
@engine_properties_options(create_mode=False)
@option(
    "--new-engine-name",
    help="Set this parameter for renaming the engine.",
    default=None,
    required=False,
)
@json_option
@exit_on_firebolt_exception
def update(
    use_spot: Optional[bool], auto_stop: int, scale: int, **raw_config_options: str
) -> None:
    """
    Update engine parameters. Engine should be stopped before updating.
    """
    something_to_update = (
        any(
            raw_config_options[param] is not None
            for param in [
                "spec",
                "type",
                "warmup",
                "description",
            ]
        )
        or scale is not None
        or use_spot is not None
        or auto_stop is not None
    )

    if not something_to_update:
        echo("Nothing to update. At least one parameter should be provided.", err=True)
        sys.exit(os.EX_USAGE)

    rm = construct_resource_manager(**raw_config_options)

    engine = rm.engines.get_by_name(name=raw_config_options["name"])

    engine = engine.update(
        name=raw_config_options["new_engine_name"],
        spec=raw_config_options["spec"],
        engine_type=ENGINE_TYPES.get(raw_config_options["type"], None),
        scale=scale,
        auto_stop=auto_stop,
        warmup=WARMUP_METHODS.get(raw_config_options["warmup"], None),
        description=raw_config_options["description"],
        use_spot=use_spot,
    )

    if not raw_config_options["json"]:
        echo(f"Engine {engine.name} is successfully updated.")

    echo_engine_information(rm, engine, bool(raw_config_options["json"]))


@command()
@common_options
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Alternatively to engine name, database name could be specified, "
    "its default engine will be used",
    hidden=True,
    callback=default_from_config_file(required=False),
)
@argument("engine_name", type=str, required=False)
@exit_on_firebolt_exception
def status(**raw_config_options: str) -> None:
    """
    Check the ENGINE_NAME status. If ENGINE_NAME is not set,
    uses default engine instead.
    """

    rm = construct_resource_manager(**raw_config_options)
    engine = get_engine_from_name_or_default(
        rm, raw_config_options["engine_name"], raw_config_options["database_name"]
    )

    current_status_name = (
        engine.current_status_summary.name if engine.current_status_summary else ""
    )
    echo(f"Engine {engine.name} current status is: {current_status_name}")


@command(name="list", short_help="List existing engines (alias: ls)")
@common_options
@option(
    "--name-contains",
    help="A string used to filter the list of returned engines. "
    "Partial matches will be returned.",
    default=None,
    type=str,
)
@option(
    "--database",
    help="Only list engines attached to this database.",
    default=None,
    type=str,
)
@json_option
@exit_on_firebolt_exception
def list(**raw_config_options: str) -> None:
    """
    List existing engines
    """

    rm = construct_resource_manager(**raw_config_options)

    if raw_config_options["database"]:
        database = rm.databases.get_by_name(raw_config_options["database"])
        engines = rm.bindings.get_engines_bound_to_database(database)

        # Filter engines by name
        if raw_config_options["name_contains"]:
            name_contains = raw_config_options["name_contains"].lower()
            engines = [
                engine
                for engine in engines
                if engine.name
                if name_contains in engine.name.lower()
            ]

        # Sort engines in alphabetical order
        engines = sorted(engines, key=lambda x: x.name)
    else:
        engines = rm.engines.get_many(
            name_contains=raw_config_options["name_contains"],
            order_by="ENGINE_ORDER_NAME_ASC",
        )

    if not raw_config_options["json"]:
        echo("Found {num_engines} engines".format(num_engines=len(engines)))

    if raw_config_options["json"] or engines:
        echo(
            prepare_execution_result_table(
                data=[
                    [
                        engine.name,
                        engine.current_status_summary.name
                        if engine.current_status_summary
                        else EngineStatusSummary.ENGINE_STATUS_SUMMARY_UNSPECIFIED,
                        rm.regions.get_by_key(engine.compute_region_key).name,
                    ]
                    for engine in engines
                ],
                header=["name", "status", "region"],
                use_json=bool(raw_config_options["json"]),
            )
        )


@command()
@common_options
@option(
    "--yes",
    help="Automatic yes on confirmation prompt",
    is_flag=True,
)
@argument(
    "engine_name",
    type=str,
)
@exit_on_firebolt_exception
def drop(**raw_config_options: str) -> None:
    """
    Drop an existing engine
    """
    rm = construct_resource_manager(**raw_config_options)
    engine = rm.engines.get_by_name(name=raw_config_options["engine_name"])

    if raw_config_options["yes"] or confirm(
        f"Do you really want to drop the engine {engine.name}?"
    ):
        engine.delete()
        echo(f"Drop request for engine {engine.name} is successfully sent")
    else:
        echo("Drop request is aborted")


@command()
@common_options
@argument(
    "engine_name",
    type=str,
)
@json_option
@exit_on_firebolt_exception
def describe(**raw_config_options: str) -> None:
    """
    Describe specified engine
    """
    rm = construct_resource_manager(**raw_config_options)
    engine = rm.engines.get_by_name(name=raw_config_options["engine_name"])
    echo_engine_information(rm, engine, bool(raw_config_options["json"]))


@command()
@common_options
@option(
    "--region",
    help="Instances information relevant to this region.",
    required=True,
    type=str,
)
@json_option
@exit_on_firebolt_exception
def get_instance_types(**raw_config_options: str) -> None:
    """
    Get instance types (spec) available for your account
    """
    rm = construct_resource_manager(**raw_config_options)
    if not raw_config_options["region"] in rm.regions.regions_by_name:
        raise FireboltError(
            f"Unknown region: {raw_config_options['region']}. "
            f"Available regions: {', '.join(rm.regions.regions_by_name.keys())}"
        )

    region = rm.regions.get_by_name(name=raw_config_options["region"])

    echo(
        prepare_execution_result_table(
            data=[
                [
                    spec.name,
                    spec.cpu_virtual_cores_count,
                    convert_bytes(spec.memory_size_bytes),
                    convert_bytes(spec.storage_size_bytes),
                ]
                for spec in sorted(
                    rm.instance_types.get_instance_types_per_region(region),
                    key=lambda x: (x.name[0], x.cpu_virtual_cores_count),
                )
            ],
            header=["name", "cpu", "memory", "storage"],
            use_json=bool(raw_config_options["json"]),
        )
    )


engine.add_command(get_instance_types)
engine.add_command(create)
engine.add_command(describe)
engine.add_command(drop)
engine.add_command(start)
engine.add_command(restart)
engine.add_command(stop)
engine.add_command(status)
engine.add_command(update)
engine.add_command(list)



import click
from click import command, echo, group, option
from firebolt_ingest.aws_settings import AWSSettings
from firebolt_ingest.table_model import Table
from firebolt_ingest.table_service import TableService

from firebolt_cli.common_options import (
    common_options,
    default_from_config_file,
)
from firebolt_cli.utils import (
    create_aws_creds_from_environ,
    create_connection,
    exit_on_firebolt_exception,
    read_from_file,
)


@group(name="table", short_help="Create tables (alias: tb)")
def table() -> None:
    """
    Create tables.
    """


@command()
@common_options
@option(
    "--engine-name",
    help="Name or url of the engine to use for SQL queries",
    envvar="FIREBOLT_ENGINE_NAME",
    callback=default_from_config_file(required=True),
)
@option(
    "--database-name",
    help="Database name to use for SQL queries",
    envvar="FIREBOLT_DATABASE_NAME",
    callback=default_from_config_file(required=True),
)
@option(
    "--s3-url",
    help="Path to the s3 bucket, where the data is stored",
    required=True,
)
@option(
    "--file",
    help="Path to the yaml config file",
    type=click.Path(exists=True),
    required=True,
)
@exit_on_firebolt_exception
def create_external(**raw_config_options: str) -> None:
    """
    Create external table
    """
    aws_settings = AWSSettings(
        s3_url=raw_config_options["s3_url"],
        aws_credentials=create_aws_creds_from_environ(),
    )

    table = Table.parse_yaml(read_from_file(raw_config_options["file"]))

    with create_connection(**raw_config_options) as connection:
        TableService(table, connection).create_external_table(aws_settings)
        echo(f"External table (ex_{table.table_name}) was successfully created")


@command()
@common_options
@option(
    "--engine-name",
    help="Name or url of the engine to use for SQL queries",
    envvar="FIREBOLT_ENGINE_NAME",
    callback=default_from_config_file(required=False),
)
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Database name to use for SQL queries",
    callback=default_from_config_file(required=True),
)
@option(
    "--file",
    help="Path to the yaml config file",
    type=click.Path(exists=True),
    required=True,
)
@option(
    "--add-file-metadata",
    help="Add meta columns (source_file_name and source_file_timestamp)"
    " to the fact table",
    is_flag=True,
    default=False,
)
@exit_on_firebolt_exception
def create_fact(**raw_config_options: str) -> None:
    """
    Create fact table
    """
    table = Table.parse_yaml(read_from_file(raw_config_options["file"]))

    with create_connection(**raw_config_options) as connection:
        TableService(table, connection).create_internal_table(
            add_file_metadata=raw_config_options["add_file_metadata"]
        )
        echo(f"Fact table ({table.table_name}) was successfully created")


table.add_command(create_external)
table.add_command(create_fact)



import csv
import os
import sys
import time
from typing import Optional

import click
import sqlparse  # type: ignore
from click import command, echo, option
from firebolt.async_db.cursor import Statistics
from firebolt.common.exception import FireboltError
from firebolt.db import Connection, Cursor
from prompt_toolkit.application import get_app
from prompt_toolkit.completion import DynamicCompleter, ThreadedCompleter
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Condition
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.shortcuts import PromptSession
from pygments.lexers import PostgresLexer
from tabulate import tabulate

from firebolt_cli.common_options import (
    common_options,
    default_from_config_file,
)
from firebolt_cli.completer import FireboltAutoCompleter
from firebolt_cli.utils import (
    construct_resource_manager,
    convert_bytes,
    convert_num_human_readable,
    create_connection,
    exit_on_firebolt_exception,
    format_short_statement,
    get_default_database_engine,
    read_from_file,
    read_from_stdin_buffer,
)

EXIT_COMMANDS = [".exit", ".quit", ".q"]
HELP_COMMANDS = [".help", ".h"]
TABLES_COMMAND = ".tables"
INTERNAL_COMMANDS = EXIT_COMMANDS + HELP_COMMANDS + [TABLES_COMMAND]


def is_data_statement(statement: sqlparse.sql.Statement) -> bool:
    """
    true if the statement is supposed to output some data
    """
    token = statement.token_first(skip_cm=True, skip_ws=True).normalized
    return token in {"SHOW", "DESCRIBE", "EXPLAIN", "SELECT", "WITH"}


def format_time(execution_time: float) -> str:
    """
    format time from seconds into string in format: 99h 59m 59.99s
    """
    hours = int(execution_time // 3600)
    execution_time -= hours * 3600

    minutes = int(execution_time // 60)
    execution_time -= minutes * 60

    return (
        (f"{hours}h " if hours > 0 else "")
        + (f"{minutes}m " if minutes > 0 or hours > 0 else "")
        + f"{execution_time:.2f}s"
    )


def format_statistics(statistics: Statistics, execution_time: float) -> str:
    """
    format statistics from the query execution request to a string
    Args:
        statistics: query execution statistics
        execution_time: real execution time

    Returns:
        a formatted string
    """
    # Calculate cached data ratio
    cached_data_ratio = None
    if (
        statistics.scanned_bytes_cache is not None
        and statistics.scanned_bytes_storage is not None
    ):
        scanned_bytes_total = (
            statistics.scanned_bytes_cache + statistics.scanned_bytes_storage
        )
        cached_data_ratio = (
            statistics.scanned_bytes_cache / scanned_bytes_total
            if scanned_bytes_total
            else 1
        )
    rows_per_second = statistics.rows_read / statistics.elapsed

    return (
        f"\nTotal elapsed time    : {format_time(execution_time)}"
        f"\nFirebolt elapsed time : {format_time(statistics.elapsed)}"
        f"\nScanned bytes         : {convert_bytes(statistics.bytes_read)}"
        f"\nRows / Second         : {convert_num_human_readable(rows_per_second)}"
        + (
            f"\nCached data ratio     : {cached_data_ratio * 100:.2f}%"
            if cached_data_ratio
            else ""
        )
    )


def echo_execution_status(
    statement: str,
    statement_idx: int,
    statements_num: int,
    execution_time: float,
    success: bool,
    statistics: Optional[Statistics],
) -> None:
    """
    print execution summary result: Success or Error
    and a shortened statement to which it is referring
    """
    statement = format_short_statement(statement)
    counter = "" if statements_num < 2 else f"({statement_idx}/{statements_num}) "

    statistics_message = (
        format_statistics(statistics, execution_time) if statistics else ""
    )

    msg, color = (
        (f"{counter}Success", "green") if success else (f"{counter}Error", "yellow")
    )

    echo(
        "{} ({}) {}".format(
            click.style(msg, fg=color, bold=True), statement, statistics_message
        )
    )


def execute_and_print(cursor: Cursor, query: str, use_csv: bool) -> None:
    """
    Execute multiple queries one by one, fetch the data from cursor
    and print it in csv or tabular format.
    """
    statements = sqlparse.parse(query)
    for statement_idx, statement in enumerate(statements):
        try:
            start_time = time.time()
            cursor.execute(str(statement))
            execution_time = time.time() - start_time

            is_data = is_data_statement(statement)

            if not use_csv:
                echo_execution_status(
                    str(statement),
                    statement_idx + 1,
                    len(statements),
                    execution_time,
                    success=True,
                    statistics=cursor.statistics,
                )

            if cursor.description and is_data:
                data = cursor.fetchall()

                headers = [i.name for i in cursor.description]
                if use_csv:
                    writer = csv.writer(sys.stdout)
                    writer.writerow(headers)
                    writer.writerows(data)
                else:
                    echo(
                        tabulate(
                            data,
                            headers=headers,
                            tablefmt="github",
                            disable_numparse=True,
                        )
                    )

            cursor.nextset()

        except FireboltError as err:
            execution_time = time.time() - start_time
            echo_execution_status(
                str(statement),
                statement_idx + 1,
                len(statements),
                execution_time,
                success=False,
                statistics=cursor.statistics,
            )
            raise err


@Condition
def is_multiline_needed() -> bool:
    """
    function reads the buffer of the interactive prompt
    and return true if the continuation of the request is required
    """
    buffer = get_app().layout.get_buffer_by_name(DEFAULT_BUFFER)
    if buffer is None:
        return True

    text = buffer.text.strip()
    return not (len(text) == 0 or text.endswith(";") or text in INTERNAL_COMMANDS)


def show_help() -> None:
    """
    Print help message with internal commands of interactive sql execution
    """
    rows = [
        ["/".join(HELP_COMMANDS), "Show this help message."],
        ["/".join(EXIT_COMMANDS), "Exit firebolt-cli."],
        [TABLES_COMMAND, "Show tables in current database."],
    ]

    for internal_command, help_message in rows:
        echo("{:<15s}".format(internal_command), nl=False)
        echo(help_message)


def process_internal_command(internal_command: str) -> str:
    """
    process internal command, execute an internal command
    or make an sql query from internal_command
    """
    if internal_command in EXIT_COMMANDS:
        raise EOFError()
    elif internal_command in HELP_COMMANDS:
        show_help()
        return ""
    elif internal_command == TABLES_COMMAND:
        return "SHOW tables;"

    raise ValueError(f"Not known internal command: {internal_command}")


def enter_interactive_session(connection: Connection, use_csv: bool) -> None:
    """
    Enters an infinite loop of interactive shell.
    """
    echo("Connection succeeded.")

    completer = FireboltAutoCompleter(connection.cursor())

    session: PromptSession = PromptSession(
        message="firebolt> ",
        prompt_continuation="     ...> ",
        lexer=PygmentsLexer(PostgresLexer),
        completer=ThreadedCompleter(DynamicCompleter(lambda: completer)),
        multiline=is_multiline_needed,
    )

    cursor = connection.cursor()
    while 1:
        try:
            sql_query = session.prompt()
            sql_query = sql_query.strip().rstrip(";")

            if sql_query in INTERNAL_COMMANDS:
                sql_query = process_internal_command(sql_query)

            if len(sql_query) == 0:
                continue

            execute_and_print(cursor, sql_query, use_csv)
        except FireboltError as err:
            echo(err)
            continue
        except KeyboardInterrupt:
            continue
        except EOFError:
            echo("Bye!")
            break


@command()
@common_options
@option(
    "--engine-name",
    help="Name or url of the engine to use for SQL queries",
    envvar="FIREBOLT_ENGINE_NAME",
    callback=default_from_config_file(required=False),
)
@option(
    "--csv", help="Provide query output in CSV format.", is_flag=True, default=False
)
@option(
    "--database-name",
    envvar="FIREBOLT_DATABASE_NAME",
    help="Database name to use for SQL queries.",
    callback=default_from_config_file(),
)
@option(
    "--file",
    help="Path to the file with the SQL query to be executed.",
    default=None,
    type=click.Path(exists=True),
)
@option("--sql", help="SQL statement, that will be executed", required=False)
@exit_on_firebolt_exception
def query(**raw_config_options: str) -> None:
    """
    Execute SQL queries.
    """
    stdin_query = read_from_stdin_buffer()
    file_query = read_from_file(raw_config_options["file"])
    args_query = raw_config_options["sql"]

    if bool(stdin_query) + bool(file_query) + bool(args_query) > 1:
        echo(
            "SQL request should be either read from stdin or file or "
            "command line arguments. Multiple are specified.",
            err=True,
        )
        sys.exit(os.EX_USAGE)

    sql_query = stdin_query or file_query or args_query

    # if engine_name is not set, use default engine
    if raw_config_options["engine_name"] is None:
        rm = construct_resource_manager(**raw_config_options)
        raw_config_options["engine_name"] = get_default_database_engine(
            rm, raw_config_options["database_name"]
        ).endpoint

    with create_connection(**raw_config_options) as connection:

        if sql_query:
            # if query is available, then execute, print result and exit
            execute_and_print(
                connection.cursor(), sql_query, bool(raw_config_options["csv"])
            )
        else:
            # otherwise start the interactive session
            enter_interactive_session(connection, bool(raw_config_options["csv"]))



KEYWORDS = [
    "ALTER DATABASE",
    "ALTER ENGINE",
    "ALTER TABLE",
    "AND",
    "AS",
    "AS SELECT",
    "ASC",
    "ATTACH ENGINE",
    "AUTO_STOP",
    "CASCADE",
    "CASE",
    "COMPRESSION",
    "COPY",
    "CREATE AGGREGATING INDEX",
    "CREATE AND GENERATE AGGREGATING INDEX",
    "CREATE DATABASE",
    "CREATE DIMENSION TABLE",
    "CREATE ENGINE",
    "CREATE EXTERNAL TABLE",
    "CREATE FACT TABLE",
    "CREATE JOIN INDEX",
    "CREATE VIEW",
    "CREDENTIALS",
    "CSV",
    "DEFAULT",
    "DESC",
    "DESCRIBE",
    "DROP",
    "DROP AGGREGATING INDEX",
    "DROP DATABASE",
    "DROP ENGINE",
    "DROP INDEX",
    "DROP JOIN INDEX",
    "DROP PARTITION",
    "DROP TABLE",
    "DROP VIEW",
    "ELSE",
    "EXPLAIN",
    "EXPLAIN USING JSON",
    "EXPLAIN USING TEXT",
    "FALSE",
    "FILE_NAME_PREFIX",
    "FROM",
    "GROUP BY",
    "GZIP",
    "HAVING",
    "IF",
    "IF EXISTS",
    "IF NOT EXISTS",
    "IN",
    "INCLUDE_QUERY_ID_IN_FILE_NAME",
    "INSERT INTO",
    "JSON",
    "LIMIT",
    "MAX_FILE_SIZE",
    "NONE",
    "NOT NULL",
    "NULL",
    "NULLS FIRST",
    "NULLS LAST",
    "OBJECT_PATTERN",
    "OFFSET",
    "ON",
    "OR",
    "OR REPLACE",
    "ORDER BY",
    "OVERWRITE_EXISTING_FILES",
    "PARQUET",
    "PARTITION",
    "PARTITION BY",
    "PRIMARY INDEX",
    "REFRESH ALL JOIN INDEXES ON TABLE",
    "REFRESH JOIN INDEX",
    "RENAME TO",
    "SCALE",
    "SELECT",
    "SELECT DISTINCT",
    "SET",
    "SHOW CACHE",
    "SHOW COLUMNS",
    "SHOW DATABASES",
    "SHOW ENGINES",
    "SHOW INDEXES",
    "SHOW TABLES",
    "SHOW VIEWS",
    "SINGLE_FILE",
    "SPEC",
    "START ENGINE",
    "STOP ENGINE",
    "THEN",
    "TO",
    "TRUE",
    "TSV",
    "TYPE",
    "UNION",
    "UNIQUE",
    "URL",
    "VALUES",
    "WARMUP",
    "WHERE",
    "WHEN",
    "WITH",
]

FUNCTIONS = [
    "ABS(<val>)",
    "ACOS(<val>)",
    "ALL_MATCH(<array_var> -> <array_var>[<comparison>]<expr>, <array_expr>)",
    "ANY_MATCH(<func>, <arr>)",
    "ANY(<col>)",
    "ANY_VALUE(<col>)",
    "APPROX_PERCENTILE(<expr>,<percent>)",
    "ARRAY_CONCAT(<arr1> [, ...n])",
    "ARRAY_COUNT_GLOBAL(<arr_col>)",
    "ARRAY_COUNT(<func>, <arr>)",
    "ARRAY_CUMULATIVE_SUM( [<func>,] arr)",
    "ARRAY_DISTINCT(<arr>)",
    "ARRAY_FILL(<func>, <arr>)",
    "ARRAY_FIRST_INDEX(<func>, <arr>)",
    "ARRAY_FIRST(<func>, <arr>)",
    "ARRAY_INTERSECT(<arr>)",
    "ARRAY_JOIN(<arr>[, <delimiter>])",
    "ARRAY_MAX_GLOBAL(<arr>)",
    "ARRAY_MAX(<arr>)",
    "ARRAY_MIN_GLOBAL(<arr>)",
    "ARRAY_MIN(<arr>)",
    "ARRAY_REPLACE_BACKWARDS(<func>, <arr>)",
    "ARRAY_REVERSE(<arr>)",
    "ARRAY_SORT([<func>,] <arr>)",
    "ARRAY_SUM_GLOBAL(<arr>)",
    "ARRAY_SUM([<func>,] <arr>)",
    "ARRAY_UNIQ(<arr> [, ...n])",
    "ARRAY_UNNEST(<arr>)",
    "ASIN(<val>)",
    "ATAN(<expr>)",
    "ATAN2(<y_expr>,<x_expr>)",
    "AVG( <val> ) OVER ( [ PARTITION BY <exp> ] )",
    "AVG(<expr>)",
    "BASE64_ENCODE(<expr>)",
    "CAST(<value> AS <type>)",
    "CBRT(<val>)",
    "CEIL(<val>[, <dec>])",
    "CHECKSUM( <expr1> [, <expr2>] [, <expr3>] [, ...<exprN>] )",
    "CITY_HASH(<exp>, [, expr2 [,...]])",
    "COALESCE(<value> [,...])",
    "CONCAT( <string>, <string2>[, ...n] )",
    "CONTAINS(<arr>, <arg>)",
    "COS(<exp>)",
    "COT(<exp>)",
    "COUNT( <val> ) OVER ( [ PARTITION BY <exp> ] )",
    "COUNT([ DISTINCT ] <expr>)",
    "CURRENT_DATE()",
    "DATE_ADD('<unit>', <interval>, <date_expr>)",
    "DATE_DIFF('<unit>', <start_date>, <end_date>)",
    "DATE_FORMAT(<date>, '<format>')",
    "DATE_TRUNC('<precision>', <date>)",
    "DEGREES(<exp>)",
    "DENSE_RANK() OVER ([PARTITION BY <val>] ORDER BY <exp> [ASC|DESC] )",
    "ELEMENT_AT(<arr>, <index>)",
    "EXP(<val>)",
    "EXTRACT_ALL( <expr>, '<regex_pattern>' )",
    "EXTRACT(<field> FROM <source>)",
    "FILTER(<func>, <arr> [, ...n] )",
    "FLATTEN(<arr_of_arrs>)",
    "FLOOR(<val>[, <dec>])",
    "FROM_UNIXTIME(<unix_time>)",
    "GEN_RANDOM_UUID()",
    "IFNULL(<exp1>, <exp2>)",
    "INDEX_OF(<arr>, <x>)",
    "JSON_EXTRACT_ARRAY_RAW(<json>, '<json_pointer_expression>')",
    "JSON_EXTRACT_KEYS(<json>, '<json_pointer_expression>')",
    "JSON_EXTRACT_RAW(<json>, <json_pointer_expression>)",
    "JSON_EXTRACT_VALUES(<json>, '<json_pointer_expression>')",
    "JSON_EXTRACT(<json>, '<json_pointer_expression>', '<expected_type>')",
    "LAG(first_name, 1) OVER (PARTITION BY grade_level ORDER BY first_name )",
    "LEAD(first_name, -1) OVER (PARTITION BY grade_level ORDER BY first_name )",
    "LENGTH(<string>)",
    "LENGTH(<arr>)",
    "LOG([<base>,] <num>)",
    "LOWER(<string>)",
    "LPAD(<str>, <length>[, <pad>])",
    "LTRIM(<target>)",
    "MATCH_ANY(<string>, <pattern_array>)",
    "MATCH(<string>, '<pattern>')",
    "MAX_BY(<arg>, <val>)",
    "MAX( <exp> ) OVER ( [ PARTITION BY <exp> ] )",
    "MAX(<expr>)",
    "MD5_NUMBER_LOWER64(<string>)",
    "MD5_NUMBER_UPPER64(<string>)",
    "MD5(<string>)",
    "MEDIAN(<col>)",
    "MIN_BY(arg, val)",
    "MIN( <exp> ) OVER ( [ PARTITION BY <exp> ] )",
    "MIN(<expr>)",
    "MOD(<num>,<den>)",
    "NEST(<col>)",
    "NOW()",
    "NULLIF(<exp1>, <exp2>)",
    "PI()",
    "POW(<val>, <exp>)",
    "RADIANS(<val>)",
    "RANDOM()",
    "RANK() OVER ([PARTITION BY <exp>] ORDER BY <exp> [ASC|DESC] )",
    "REDUCE(<agg_function>, <arr>)",
    "REGEXP_LIKE(<string>, '<pattern>')",
    "REGEXP_MATCHES(<string>, <pattern>[,'<flag>[...]'])",
    "REPEAT(<string>, <repeating_number>)",
    "REPLACE('hello world','hello','nice')",
    "REVERSE(<string>)",
    "ROUND(<val> [, <dec>])",
    "ROW_NUMBER() OVER ([PARTITION BY <exp>] ORDER BY <exp> [ASC|DESC] )",
    "RPAD(<str>, <length>[, <pad>])",
    "RTRIM(<target>)",
    "SIGN(<expr>)",
    "SIN(<val>)",
    "SLICE(<arr>, <offset>[, <length>])",
    "SPLIT_PART(<string>, <delimiter>, <index>)",
    "SPLIT( <delimiter>, <string> )",
    "SQRT(<val>)",
    "STDDEV_SAMP(<expr>)",
    "STRPOS(<string>, <substring>)",
    "SUBSTR(<string>, <offset> [, <length>])",
    "SUM([DISTINCT] <val> ) OVER ( [ PARTITION BY <expr> ] )",
    "SUM(score)",
    "TAN(<val>)",
    "TIMEZONE()",
    "TO_CHAR(<expr>[, '<format>'])",
    "TO_DATE(<string>)",
    "TO_DAY_OF_WEEK(<date>)",
    "TO_DAY_OF_YEAR(<date>)",
    "TO_DOUBLE(<exp>)",
    "TO_FLOAT(<expr>)",
    "TO_HOUR(<timestamp>)",
    "TO_INT(<exp>)",
    "TO_LONG(<exp>)",
    "TO_MINUTE(<timestamp>)",
    "TO_MONTH(<date>)",
    "TO_QUARTER(<date>)",
    "TO_SECOND(<timestamp>)",
    "TO_STRING(<date>)",
    "TO_TIMESTAMP(<string>)",
    "TO_UNIX_TIMESTAMP(<string>)",
    "TO_UNIXTIME(<string>)",
    "TO_WEEK(<date>)",
    "TO_WEEKISO(<date>)",
    "TO_YEAR(<date>)",
    "TRANSFORM(<func>, <arr>)",
    "TRIM( [LEADING | TRAILING | BOTH] <trim_character> FROM <target_string>)",
    "TRUNC(<val>[, <dec>])",
    "TRY_CAST(<value> AS <type>)",
    "UPPER(<string>)",
    "ZEROIFNULL(<expr>)",
]



import json
import os
import re
import sys
from configparser import ConfigParser
from functools import lru_cache, wraps
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Type

import keyring
import sqlparse  # type: ignore
from appdirs import user_config_dir
from click import Command, Context, Group, echo
from firebolt.client.auth import Token, UsernamePassword
from firebolt.common import Settings
from firebolt.common.exception import FireboltError
from firebolt.db.connection import Connection, connect
from firebolt.model.engine import Engine
from firebolt.service.manager import ResourceManager
from firebolt_ingest.aws_settings import (
    AWSCredentials,
    AWSCredentialsKeySecret,
    AWSCredentialsRole,
)
from httpx import HTTPStatusError
from keyring.errors import KeyringError
from tabulate import tabulate

config_file = os.path.join(user_config_dir(), "firebolt.ini")
config_section = "firebolt-cli"


def construct_shortcuts(shortages: dict) -> Type[Group]:
    class AliasedGroup(Group):
        def get_command(self, ctx: Context, cmd_name: str) -> Optional[Command]:
            rv = Group.get_command(self, ctx, cmd_name)
            if rv is not None:
                return rv

            matches = [
                x for x in self.list_commands(ctx) if x in shortages.get(cmd_name, [])
            ]

            if not matches:
                return None

            assert len(matches) == 1
            return Group.get_command(self, ctx, matches[0])

    return AliasedGroup


def format_short_statement(statement: str, truncate_long_string: int = 80) -> str:
    """
    Format a complex query into a single line with
    stripped comments and excessive whitespaces

    Args:
        statement: a valid sql statement
        truncate_long_string: if set to positive integer strings longer, that
        the specified value will be truncated and "..." will be added

    Returns: a formatted string

    """
    statement = sqlparse.format(
        str(statement), strip_comments=True, use_space_around_operators=True
    )

    statement = statement.replace("\n", " ").replace("\t", " ").strip()

    # strip consecutive whitespaces
    statement = re.sub(" +", " ", statement)

    if 0 < truncate_long_string < len(statement):
        return statement[:truncate_long_string] + " ..."

    return statement


def prepare_execution_result_line(
    data: Sequence, header: Sequence, use_json: bool = False
) -> str:
    """
    return the string representation of data in either json or tabular formats.
    In case of json, the result is dict
    In case of tabular, the result is table with headers in the first column
    """

    if len(data) != len(header):
        raise ValueError("Data and header have different length.")

    if use_json:
        return json.dumps(dict(zip(header, data)), indent=4)
    else:
        return tabulate(list(zip(header, data)), tablefmt="grid")


def prepare_execution_result_table(
    data: Sequence[Sequence], header: Sequence, use_json: bool = False
) -> str:
    """
    return the string representation of data in either json or tabular formats
    In case of json, the result is list of dicts
    In case of tabular, the result is table with headers in the first row
    """
    for d in data:
        if len(d) != len(header):
            raise ValueError("Data and header have different length.")

    if use_json:
        return json.dumps([dict(zip(header, d)) for d in data], indent=4)
    else:
        return tabulate(data, headers=header, tablefmt="grid")


def construct_resource_manager(**raw_config_options: str) -> ResourceManager:
    """
    Propagate raw_config_options to the settings and construct a resource manager
    :rtype: object
    """
    account_name = raw_config_options.get("account_name", None)
    if account_name is not None:
        account_name = account_name.lower()

    settings_dict = {
        "server": raw_config_options["api_endpoint"],
        "default_region": raw_config_options.get("region", ""),
        "account_name": account_name,
        "user": None,
        "password": None,
    }

    if raw_config_options["access_token"] is not None:
        try:
            return ResourceManager(
                Settings(
                    **settings_dict,
                    auth=Token(raw_config_options["access_token"]),
                )
            )
        except HTTPStatusError:
            pass

    return ResourceManager(
        Settings(
            **settings_dict,
            auth=UsernamePassword(
                raw_config_options["username"], raw_config_options["password"]
            ),
        )
    )


def to_human_readable(num: Optional[float], step_unit: int, labels: List[str]) -> str:
    """
    converts a long number to a short human-readable string representation
    e.g. 1233212 -> 1.2 M

    Args:
        num: number to be converted
        step_unit: threshold for moving to the next label
        labels: labels, that should be used for each next step

    """
    if num is None:
        return ""

    if num < 0:
        raise ValueError("Value cannot be negative")

    def format_output(bytes: float, dim: str) -> str:
        return "{} {}".format(f"{bytes:.2f}".rstrip("0").rstrip("."), dim)

    for x in labels:
        num /= step_unit
        if num < step_unit:
            return format_output(num, x)

    return format_output(num, x)


def convert_bytes(num: Optional[float]) -> str:
    """
    this function will convert bytes to KB, MB, GB, TB, PB, EB, ZB, YB
    """
    return to_human_readable(
        num, step_unit=1024, labels=["KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    )


def convert_num_human_readable(num: Optional[float]) -> str:
    """
    this function will convert number to human-readable form
    """
    return to_human_readable(num, step_unit=1000, labels=["K", "M", "G", "T"])


def read_from_file(fpath: Optional[str]) -> Optional[str]:
    """
    read from file, if fpath is not None, otherwise return empty string
    """
    if fpath is None:
        return None

    with open(fpath, "r") as f:
        return f.read() or None


def read_from_stdin_buffer() -> Optional[str]:
    """
    read from buffer if stdin file descriptor is open, otherwise return empty string
    """
    if sys.stdin.isatty():
        return None

    return sys.stdin.buffer.read().decode("utf-8") or None


@lru_cache()
def read_config() -> Dict[str, str]:
    """
    :return: dict with parameters from config file, or empty dict if no parameters found
    """
    config_dict: Dict[str, Optional[str]] = {}

    config = ConfigParser(interpolation=None)
    if os.path.exists(config_file):
        config.read(config_file)
        if config.has_section(config_section):
            config_dict = dict((k, v) for k, v in config[config_section].items())

    try:
        value = keyring.get_password("firebolt-cli", "password")
        if value and len(value) != 0:
            config_dict["password"] = value
    except KeyringError:
        pass

    return dict({(k, v) for k, v in config_dict.items() if v and len(v)})


def set_keyring_param(param: str, value: str) -> bool:
    """
    Set keyring param to value, if value is an empty string, delete the param

    :return: True if operation was successful
    """

    try:
        if value == "":
            keyring.delete_password("firebolt-cli", param)
        else:
            keyring.set_password("firebolt-cli", param, value)
    except KeyringError:
        return False

    return True


def update_config(**kwargs: str) -> None:
    """
    Update the config file (or use the keyring for updating password)
    if a parameter set to None, the parameter will not be updates
    To delete the parameter, it should be set to empty string

    :param kwargs:
    :return:
    """

    # Try to update password in keyring first, and only if failed in config
    if (
        "password" in kwargs
        and kwargs["password"] is not None
        and set_keyring_param("password", kwargs["password"])
    ):
        del kwargs["password"]

    if len(kwargs):
        config = ConfigParser(interpolation=None)
        if os.path.exists(config_file):
            config.read(config_file)

        if config.has_section(config_section):
            config[config_section].update(**kwargs)
        else:
            config[config_section] = kwargs

        with open(config_file, "w") as cf:
            config.write(cf)

        read_config.cache_clear()


def exit_on_firebolt_exception(func: Callable) -> Callable:
    """
    Decorator which catches all Exceptions and exits the programms
    """

    @wraps(func)
    def decorator(*args: str, **kwargs: str) -> None:
        try:
            func(*args, **kwargs)
        except Exception as err:
            echo(err, err=True)
            sys.exit(1)

    return decorator


def get_default_database_engine(rm: ResourceManager, database_name: str) -> Engine:
    """
    Get the default engine of the database. If the default engine doesn't exists
    raise FireboltError
    """

    database = rm.databases.get_by_name(name=database_name)
    bindings = rm.bindings.get_many(database_id=database.database_id)

    if len(bindings) == 0:
        raise FireboltError("No engines attached to the database")

    for binding in bindings:
        if binding.is_default_engine:
            return rm.engines.get(binding.engine_id)

    raise FireboltError("No default engine is found.")


def extract_engine_name_url(
    engine_name_url: str,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Returns a tuple engine_name, engine_url
    """
    if "." in engine_name_url:
        return None, engine_name_url
    else:
        return engine_name_url, None


def create_connection(
    engine_name: str,
    database_name: str,
    username: str,
    password: str,
    access_token: Optional[str],
    api_endpoint: str,
    account_name: Optional[str],
    **kwargs: str,
) -> Connection:
    """
    Create connection based on access_token if provided,
    in case of failure use username/password
    """
    if account_name is not None:
        account_name = account_name.lower()

    params = {
        "database": database_name,
        "api_endpoint": api_endpoint,
        "account_name": account_name,
    }

    # decide what to propagate engine_name or url
    params["engine_name"], params["engine_url"] = extract_engine_name_url(engine_name)

    if access_token:
        try:
            return connect(**params, auth=Token(access_token))
        except FireboltError:
            pass

    return connect(**params, auth=UsernamePassword(username, password))


def create_aws_key_secret_creds_from_environ() -> Optional[AWSCredentialsKeySecret]:
    """
    if FIREBOLT_AWS_KEY_ID/FIREBOLT_AWS_SECRET_KEY are set,
    construct AWSCredentialsKeySecret based on these variable.
    if both parameter are not set, returns None
    If only one parameter is set, raises an exception
    """
    aws_key_id = os.environ.get("FIREBOLT_AWS_KEY_ID")
    aws_secret_key = os.environ.get("FIREBOLT_AWS_SECRET_KEY")

    if aws_key_id and aws_secret_key:
        return AWSCredentialsKeySecret(
            aws_key_id=aws_key_id, aws_secret_key=aws_secret_key
        )
    elif aws_key_id or aws_secret_key:
        raise FireboltError(
            "Aws key/secret are both mandatory for a valid pair."
            "Provided only one parameter."
        )
    else:
        return None


def create_aws_role_creds_from_environ() -> Optional[AWSCredentialsRole]:
    """
    if FIREBOLT_AWS_ROLE_ARN is set returns AWSCredentialsRole from it and
    optionally from FIREBOLT_AWS_ROLE_EXTERNAL_ID
    if only FIREBOLT_AWS_ROLE_EXTERNAL_ID is set, raises an error
    """
    role_arn = os.environ.get("FIREBOLT_AWS_ROLE_ARN")
    external_id = os.environ.get("FIREBOLT_AWS_ROLE_EXTERNAL_ID")

    if role_arn:
        return AWSCredentialsRole(role_arn=role_arn, external_id=external_id)
    elif external_id:
        raise FireboltError("Aws external id is provided, but not role_arn")
    else:
        return None


def create_aws_creds_from_environ() -> Optional[AWSCredentials]:
    """
    Returns: AWSCredentials constructed from the provided environment variables;
        either from FIREBOLT_AWS_KEY_ID/FIREBOLT_AWS_SECRET_KEY pair
        or from FIREBOLT_AWS_ROLE_ARN/FIREBOLT_AWS_ROLE_EXTERNAL_ID
        in case of inconsistency raises FireboltError
    """

    key_secret_creds = create_aws_key_secret_creds_from_environ()
    role_creds = create_aws_role_creds_from_environ()

    if key_secret_creds is None and role_creds is None:
        return None

    if key_secret_creds and role_creds:
        raise FireboltError(
            "Either aws key/secret or role_arn/external_id pair "
            "should be specified. Found both."
        )

    return AWSCredentials(key_secret_creds=key_secret_creds, role_creds=role_creds)



import re
from logging import getLogger
from typing import Dict, Iterator, List

from firebolt.common.exception import FireboltError
from firebolt.db import Cursor
from httpx import HTTPStatusError
from prompt_toolkit import HTML
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document

from firebolt_cli.keywords import FUNCTIONS, KEYWORDS

logger = getLogger(__name__)


def prepare_display_html(text: str, offset: int) -> HTML:
    """
    prepares a colored html from the text
    """
    return HTML(
        f"<b><style color='red'>{text[:offset]}</style></b>"
        f"{text[offset:].replace('<', '&#60;').replace('>', '&#62;')}"
    )


def extract_last_word(text: str) -> str:
    """
    return the last word from the string
    """
    words = re.split("\\ |,|\n|\\)|;|\\(|\\.", text)
    return words[-1] if words else ""


def extract_last_complete_word(text: str) -> str:
    """
    extract last complete word (assuming that the last word could incomplete,
    return the word before it)
    """
    words = re.split("\\ |\n|;", text)
    return extract_last_word(words[-2]) if len(words) > 1 else ""


class FireboltAutoCompleter(Completer):
    """
    Implements autocompletion for firebolt cli
    """

    def __init__(self, cursor: Cursor):
        """
        Args:
            cursor: Cursor for executing queries and getting table and column names
        """
        self.table_columns_mapping: Dict[str, List] = {}
        self.suggestions: List = []
        self.suggestions.extend((keyword, "KEYWORD") for keyword in KEYWORDS)
        self.suggestions.extend((function, "FUNCTION") for function in FUNCTIONS)
        self.set_statements: List = []

        self.populate_table_and_column_names(cursor)
        self.populate_set_statemets(cursor)

    def populate_table_and_column_names(self, cursor: Cursor) -> None:
        """
        populate suggestion with table and column names
        """
        try:
            cursor.execute(
                "SELECT table_name, column_name, data_type "
                "FROM information_schema.columns"
            )

            data = cursor.fetchall()
            for tb_name, col_name, dtype in data:
                tb_name = str(tb_name)
                if tb_name not in self.table_columns_mapping:
                    self.table_columns_mapping[tb_name] = []

                self.table_columns_mapping[tb_name].append((col_name, dtype))

            self.suggestions.extend(
                (table_name, "TABLE")
                for table_name in self.table_columns_mapping.keys()
            )

        except (FireboltError, HTTPStatusError) as e:
            logger.info(
                f"Extraction of the list of table "
                f"and columns names failed with: {str(e)}"
            )

    def populate_set_statemets(self, cursor: Cursor) -> None:
        """
        fetch all available set statements parameters
        """
        try:
            cursor.execute("set use_standard_sql = 0")
            cursor.execute("SELECT name FROM system.settings")
            data = cursor.fetchall()

            self.set_statements.extend(
                (set_statement[0], "SET KEYWORD") for set_statement in data
            )

        except (FireboltError, HTTPStatusError) as e:
            logger.info(
                f"Extraction of the list of available "
                f"set parameters failed with: {str(e)}"
            )

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterator[Completion]:
        """
        Returns: a list of completions based on the document.text.
        Supports autocompletion by:
            - keywords
            - function names
            - table and columns names
        """
        last_word = extract_last_word(document.text_before_cursor).upper()
        last_full_word = extract_last_complete_word(document.text_before_cursor).upper()

        current_suggestions: List = []
        if last_full_word == "SET":
            current_suggestions.extend(self.set_statements)
        elif len(last_word) != 0:
            current_suggestions.extend(self.suggestions)

            for tb_name, columns in self.table_columns_mapping.items():
                if tb_name in document.text:
                    current_suggestions.extend(
                        (column_name, f"COLUMN ({column_type}, {tb_name})")
                        for column_name, column_type in columns
                    )

        offset = len(last_word)
        for label, meta in current_suggestions:
            if not label.upper().startswith(last_word):
                continue

            yield Completion(
                label,
                start_position=-offset,
                display=prepare_display_html(label, offset),
                display_meta=meta,
            )



from os import environ
from typing import Callable, List, Optional

from click import Context, MissingParameter, Parameter, option, prompt
from firebolt.client import DEFAULT_API_URL

from firebolt_cli.utils import read_config


def default_from_config_file(
    default: Optional[str] = None, required: bool = True
) -> Callable:
    def inner(ctx: Context, param: Parameter, value: Optional[str]) -> Optional[str]:
        # type check
        assert param.name

        value = value or read_config().get(param.name, None) or default
        if required and not value:
            raise MissingParameter(
                ctx=ctx,
                param=param,
                param_hint="--{}".format(param.name.replace("_", "-")),
            )
        return value

    return inner


def password_from_config_file(
    ctx: Context, param: Parameter, value: bool
) -> Optional[str]:
    # type check
    assert param.name

    # user asked to prompt for password
    if value:
        return prompt("Password", type=str, hide_input=True)

    pw_value = environ.get("FIREBOLT_PASSWORD") or read_config().get("password", None)
    if not pw_value:
        raise MissingParameter(
            ctx=ctx, param=param, param_hint="--{}".format(param.name.replace("_", "-"))
        )

    return pw_value


_common_options: List[Callable] = [
    option(
        "-u",
        "--username",
        envvar="FIREBOLT_USERNAME",
        callback=default_from_config_file(required=True),
        help="The username used for connecting to Firebolt.",
    ),
    option(
        "-p",
        "--password",
        is_flag=True,
        callback=password_from_config_file,
        help=" The password used for connecting to Firebolt.",
    ),
    option(
        "--account-name",
        envvar="FIREBOLT_ACCOUNT_NAME",
        callback=default_from_config_file(required=False),
        help="The name of the Firebolt account.",
    ),
    option(
        "--api-endpoint",
        envvar="FIREBOLT_API_ENDPOINT",
        callback=default_from_config_file(DEFAULT_API_URL, required=False),
        hidden=True,
    ),
    option(
        "--access-token",
        envvar="FIREBOLT_ACCESS_TOKEN",
        help="Firebolt token for authentication. "
        "If the access-token fails, the username/password will be used instead.",
        required=False,
    ),
]


def common_options(command: Callable) -> Callable:
    for add_option in reversed(_common_options):
        command = add_option(command)
    return command


def json_option(command: Callable) -> Callable:
    return option(
        "--json",
        help="Provide output in JSON format.",
        default=False,
        is_flag=True,
        multiple=False,
    )(command)



from click import argument, command, confirm, echo, group, option
from firebolt.model.database import Database
from firebolt.service.manager import ResourceManager

from firebolt_cli.common_options import common_options, json_option
from firebolt_cli.utils import (
    construct_resource_manager,
    construct_shortcuts,
    convert_bytes,
    exit_on_firebolt_exception,
    prepare_execution_result_line,
    prepare_execution_result_table,
)


def print_db_full_information(
    rm: ResourceManager, database: Database, use_json: bool
) -> None:
    attached_engines = rm.bindings.get_engines_bound_to_database(database)
    attached_engine_names = [str(engine.name) for engine in attached_engines]

    echo(
        prepare_execution_result_line(
            data=[
                database.name,
                database.description,
                str(rm.regions.get_by_key(database.compute_region_key).name),
                convert_bytes(database.data_size_full),
                str(database.create_time),
                attached_engine_names,
            ],
            header=[
                "name",
                "description",
                "region",
                "data_size",
                "create_time",
                "attached_engine_names",
            ],
            use_json=use_json,
        )
    )


@group(
    name="database",
    cls=construct_shortcuts(
        shortages={
            "list": "list (ls)",
            "ls": "list (ls)",
        }
    ),
    short_help="Manage the databases (alias: db)",
)
def database() -> None:
    """
    Manage databases.
    """


@command()
@common_options
@option("--name", help="A new database name.", type=str, required=True)
@option(
    "--description",
    help="A database description (max: 64 characters).",
    type=str,
    default="",
)
@json_option
@option("--region", help="The region for the new database.", required=True, type=str)
@exit_on_firebolt_exception
def create(**raw_config_options: str) -> None:
    """
    Create a new database
    """

    rm = construct_resource_manager(**raw_config_options)

    database = rm.databases.create(
        name=raw_config_options["name"],
        description=raw_config_options["description"],
        region=raw_config_options["region"],
    )

    if not raw_config_options["json"]:
        echo(f"Database {database.name} is successfully created")

    print_db_full_information(rm, database, bool(raw_config_options["json"]))


@command(name="list", short_help="List existing databases (alias: ls)")
@common_options
@option(
    "--name-contains",
    help="A string used to filter the list of returned databases. "
    "Partial matches will be returned. ",
    default=None,
    type=str,
)
@json_option
@exit_on_firebolt_exception
def list(**raw_config_options: str) -> None:
    """
    List existing databases
    """
    rm = construct_resource_manager(**raw_config_options)

    databases = rm.databases.get_many(
        name_contains=raw_config_options["name_contains"],
        order_by="DATABASE_ORDER_NAME_ASC",
    )

    if not raw_config_options["json"]:
        echo(f"Found {len(databases)} databases")

    if raw_config_options["json"] or databases:
        echo(
            prepare_execution_result_table(
                data=[
                    [
                        db.name,
                        str(rm.regions.get_by_key(db.compute_region_key).name),
                        db.description,
                    ]
                    for db in databases
                ],
                header=["name", "region", "description"],
                use_json=bool(raw_config_options["json"]),
            )
        )


@command()
@common_options
@option(
    "--yes",
    help="Automatically drop database without a confirmation prompt.",
    is_flag=True,
)
@argument(
    "database_name",
    type=str,
)
@exit_on_firebolt_exception
def drop(**raw_config_options: str) -> None:
    """
    Drop specified database. By default, this requires confirmation to execute.
    """
    rm = construct_resource_manager(**raw_config_options)
    database = rm.databases.get_by_name(name=raw_config_options["database_name"])

    if raw_config_options["yes"] or confirm(
        f"Do you really want to drop the database {database.name}?"
    ):
        database.delete()
        echo(f"Drop request for database {database.name} is successfully sent")
    else:
        echo("Drop request is aborted")


@command()
@common_options
@argument(
    "database_name",
    type=str,
)
@json_option
@exit_on_firebolt_exception
def describe(**raw_config_options: str) -> None:
    """
    Describe specified database.
    """
    rm = construct_resource_manager(**raw_config_options)
    database = rm.databases.get_by_name(name=raw_config_options["database_name"])
    print_db_full_information(rm, database, bool(raw_config_options["json"]))


@command()
@common_options
@option(
    "--name",
    help="The database name that should be updated.",
    required=True,
    type=str,
)
@option(
    "--description",
    help="Database textual description (max: 64 characters)",
    type=str,
    required=True,
)
@json_option
@exit_on_firebolt_exception
def update(**raw_config_options: str) -> None:
    """
    Update specified database description
    """
    rm = construct_resource_manager(**raw_config_options)
    database = rm.databases.get_by_name(name=raw_config_options["name"])

    database = database.update(description=raw_config_options["description"])

    if not raw_config_options["json"]:
        echo(f"The database {database.name} was successfully updated")

    print_db_full_information(rm, database, bool(raw_config_options["json"]))


database.add_command(create)
database.add_command(list)
database.add_command(drop)
database.add_command(describe)
database.add_command(update)



from click import Context, echo, group, option, pass_context, prompt

from firebolt_cli.utils import (
    exit_on_firebolt_exception,
    read_config,
    update_config,
)


@group(
    name="configure",
    short_help="Store firebolt configuration (alias: config)",
    invoke_without_command=True,
)
@option("-u", "--username", help="The email address used for connecting to Firebolt.")
@option(
    "-p",
    "--password",
    is_flag=True,
    default=False,
    help="Opens a protected prompt to enter the Firebolt password.",
)
@option("--account-name", help="The name of the Firebolt account.")
@option(
    "--database-name", help="The name of the database you would like to connect to."
)
@option("--api-endpoint", hidden=True)
@option("--engine-name", help="The name or URL of the engine to use.")
@exit_on_firebolt_exception
@pass_context
def configure(ctx: Context, **raw_config_options: str) -> None:
    """
    Store firebolt configuration parameters in config file.
    """
    if ctx.invoked_subcommand is None:
        config = {k: v for k, v in raw_config_options.items() if v is not None}

        if config["password"]:
            config["password"] = prompt("Password", type=str, hide_input=True)
        else:
            del config["password"]

        if len(config) == 0:
            prev_config = read_config()

            keys = (
                "username",
                "password",
                "account_name",
                "database_name",
                "engine_name",
            )
            keys_readable = (
                "Username",
                "Password",
                "Account name",
                "Database name",
                "Engine name or URL",
            )
            skip_message = (
                prev_config.get("username", None),
                "************" if "password" in prev_config else None,
                prev_config.get("account_name", None),
                prev_config.get("database_name", None),
                prev_config.get("engine_name", None),
            )

            for key, key_readable, message in zip(keys, keys_readable, skip_message):
                value = prompt(
                    f"{key_readable} [{message if message else None}]",
                    hide_input=key == "password",
                    default=prev_config.get(key, ""),
                    show_default=False,
                )
                config[key] = value

        if config.get("account_name", None) is not None:
            config["account_name"] = config["account_name"].lower()

        update_config(**config)
        echo("Successfully updated firebolt-cli configuration")


@configure.command()
@exit_on_firebolt_exception
def reset() -> None:
    """
    Reset all previously set configurations.
    """
    update_config(
        username="", password="", account_name="", database_name="", engine_name=""
    )
    echo("Successfully reset firebolt-cli configuration")
